module Core.Handlers where

import qualified Core.Configuration      as Conf
import           Core.Feed                  (Post(..), autoFeed, renderFeed)
import           Core.Rendering
import           Fragments.Base
import           Fragments.Styles        as S

import qualified Text.Atom.Feed          as Atom
import           Control.Monad.IO.Class     (liftIO)
import qualified Data.Text               as T
import qualified Data.Text.Lazy.Encoding as TLE
import           Data.Time.Clock            (UTCTime(..), getCurrentTime)
import           Data.Time.Format           (formatTime, defaultTimeLocale)
import           Lucid                      (Html)
import           System.Directory           (doesFileExist, getModificationTime)
import           System.FilePath.Find       ( always, extension, fileName, find, (&&?)
                                            , (/~?), (==?) )
import           System.FilePath            ( dropExtension, takeFileName )
import           Web.Twain           hiding (fileName)

-- A ResponoderM capable of lifting to IO monad; constructs response to clients
index :: ResponderM a
index = do
  (title, homeMd, footerMd) <- liftIO $ (,,)
    <$> Conf.title
    <*> mdFileToLucid "./data/posts/home.md"
    <*> mdFileToLucid "./data/posts/footer.md"
  -- Respond to request with fragments compositionally to create a home page
  sendLucidFragment $ basePage title (baseHome homeMd) footerMd

-- Responds with processed Commonmark -> HTML for posts
posts :: ResponderM a
posts = do
  postName  <- param "name"
  postValid <- liftIO $ postExists postName
  case postValid of
    False  -> missing
    True   -> do
      (title, footerMd, postMd) <- liftIO $ (,,)
        <$> Conf.title
        <*> mdFileToLucid "./data/posts/footer.md"
        <*> (mdFileToLucid $ postPath postName)
      sendLucidFragment $ basePage title (basePost postMd) footerMd
    where
      postExists :: T.Text -> IO Bool
      postExists postName = doesFileExist $ postPath postName

      postPath :: T.Text -> FilePath
      postPath postName = "./data/posts/" ++ T.unpack postName ++ ".md"

-- Builds an index of all posts
postsIndex :: ResponderM a
postsIndex = do
  (postNames, title, footerMd) <- liftIO $ (,,)
    <$> mdPostNames
    <*> Conf.title
    <*> mdFileToLucid "./data/posts/footer.md"
  sendLucidFragment $ basePage title (postIndex postNames) footerMd

-- Generates the XML feed at /feed
feed :: ResponderM a
feed = do
  (postNames, title, baseUrl, time) <- liftIO $ (,,,)
    <$> mdPostNames
    <*> Conf.title
    <*> Conf.baseUrl
    <*> fmap (\x -> timeFormat x) getCurrentTime
  feedData <- liftIO $ mapM (makePost baseUrl) postNames
  -- Send an XML response with an automatically populated Atom feed
  send $ xml $ TLE.encodeUtf8 $ renderFeed
       $ autoFeed (baseFeed title time baseUrl) feedData
    where
      -- Base feed data structure which we populate with entries
      baseFeed :: String -> String -> String -> Atom.Feed
      baseFeed title time baseUrl = Atom.nullFeed
        (T.pack $ baseUrl ++ "/feed")
        (Atom.TextString $ T.pack title)
        (T.pack $ time ++ " UTC")

      -- Create an Atom Post for each markdown post present
      makePost :: String -> FilePath -> IO (Post)
      makePost baseUrl postName = do
        date <- getModificationTime $ "./data/posts/" ++ postName ++ ".md"
        postContent <- mdFileToText $ "./data/posts/" ++ postName ++ ".md"
        return  $ Post
          (T.pack $ (timeFormat date) ++ " UTC")
          (T.pack $ baseUrl ++ "/posts/" ++ postName)
          (T.pack $ show postName)
          (postContent)

      -- YYYY-MM-DD HH:MM | 2024-02-24 16:36
      timeFormat :: UTCTime -> String
      timeFormat date = formatTime defaultTimeLocale "%Y-%m-%d %H:%M" date

-- Refer to index comments
contact :: ResponderM a
contact = do
  (title, contactMd, footerMd) <- liftIO $ (,,)
    <$> Conf.title
    <*> mdFileToLucid "./data/posts/contact.md"
    <*> mdFileToLucid "./data/posts/footer.md"
  sendLucidFragment   $ basePage title (baseContact contactMd) footerMd

-- Respond with primary processed CSS
theme :: ResponderM a
theme = send $ css $ TLE.encodeUtf8 $ S.cssRender S.composedStyles

-- Helper function for responding in ResponderM from Html
sendLucidFragment :: Html () -> ResponderM a
sendLucidFragment x = send $ html $ lucidToTwain x

-- 404 handler
missing :: ResponderM a
missing = sendLucidFragment pageNotFound

-- List of all non-hidden .md posts that aren't part of templating
mdPostNames :: IO [FilePath]
mdPostNames = mapM (pure . dropExtension . takeFileName )
                =<< find isVisible fileFilter "./data/posts"
                       where
                         isVisible  = fileName  /~? ".?*"
                         isMdFile   = extension ==? ".md"
                         isHome     = fileName  /~? "home.md"
                         isContact  = fileName  /~? "contact.md"
                         isFooter   = fileName  /~? "footer.md"
                         fileFilter = isMdFile
                                  &&? isVisible
                                  &&? isHome
                                  &&? isContact
                                  &&? isFooter
